# define context class
# written by MOROHOSHI Akihiko <moro@nii.ac.jp>

########################################################################
# Context
class Context
    attr_accessor :bold, :emphasis, :strong_emphasis,
	:italic, :typewriter
    attr_reader :underline, :size, :color
    
    SIZE_UNKNOWN = -1
    COLOR_UNKNOWN = ?U

    def initialize
	@bold = @emphasis = @strong_emphasis = @italic = @typewriter =false
	@underline = 0
	@size = SIZE_UNKNOWN
	@color = COLOR_UNKNOWN
    end

    def color=(c)
	case c
	when ?r, ?g, ?b, ?c, ?m, ?y, ?k, ?w
	    @color = c
	else
	    raise "unknown color: #{c.inspect}"
	end
    end
    def size=(s)
	if (0..9).include? s
	    @size = s
	else
	    raise "invalid size: #{s}"
	end
    end

    def underline_add
	if @underline < 3
	    @underline += 1
	end
    end

    def underline_reset
	@underline = 0
    end

    def space_context()
	c = self.dup
	c.bold = c.emphasis = c.strong_emphasis = c.italic = false
	c.color = ?w if c.underline == 0
	c
    end

    def ==(other)
	return false if other.type != self.type

	return false if other.bold != @bold
	return false if other.typewriter != @typewriter
	return false if other.italic != @italic
	return false if other.underline != @underline
	return false if other.color != @color
	return false if other.size != @size
	return false if other.strong_emphasis != @strong_emphasis
	return false if (!@strong_emphasis) and other.emphasis != @emphasis
	#return false if other.emphasis != @emphasis

	true
    end

    def eql?(other)
	self == other
    end

    def hash
	h = @bold ? 1 : 0
	h *= 3
	h += @typewriter ? 1 : 0
	h *= 3
	h += @italic ? 1 : 0
	h *= 3
	h += @underline ? 1 : 0
	h *= 3
	h += @color
	h *= 3
	h += @size
	h *= 3
	h += @strong_emphasis ? 1 : 0
	h *= 3
	h += @emphasis ? 1 : 0
	h
    end

    def distance(other)
	d = 0
	d += 1 if other.bold != @bold
	d += 1 if other.typewriter != @typewriter
	d += 1 if other.italic != @italic
	d += (other.underline - @underline).abs
	d += 1 if other.color != @color
	d += 1 if other.size != @size
	d += 1 if other.strong_emphasis != @strong_emphasis
	unless (other.strong_emphasis == @strong_emphasis and
		@strong_emphasis)
	    d += 1 if other.emphasis != @emphasis
	end
	
	d
    end

    # $B?dDj:G>.5wN%!#(Best$B$O(BESTimated.
    def distance_min_est(other)
	d = 0
	d1 = 0
	d1 += 1 if other.bold != @bold
	d1 += 1 if other.typewriter != @typewriter
	d1 += 1 if other.italic != @italic
	d1 += 1 if other.strong_emphasis != @strong_emphasis
	unless (other.strong_emphasis == @strong_emphasis and
		@strong_emphasis)
	    d1 += 1 if other.emphasis != @emphasis
	end
	
	d2 = (other.underline - @underline).abs
	
	if self.plain? or other.plain?
	    d += d1+d2
	else
	    dd1 = d1+d2
	    plain_context = context_modified_by_tag(self,'PL')
	    dd2 = self.distance_min_est_tag(plain_context) +
		plain_context.distance_min_est_tag(other)
	    d += (dd1 <= dd2 ? dd1 : dd2)
	end

	d += 1 if other.color != @color
	d += 1 if other.size != @size
	
	d
    end

    # $B?dDj:G>.5wN%(B($B%?%0$N%P%$%H?t$K$h$k(B)$B!#(B
    # $B$=$l$>$l$NCM$,(B .5 $BC10L$NC<?t$J$N$O!"(B
    # $B%H!<%/%sNs$N8e$m$+$iE,MQ$7$F$$$/$3$H$rA[Dj$7$F$$$k!#(B
    # $B>o$K3+;O%?%0$h$j$b=*N;%?%0$NJ}$,(B1$BD9$/!"(B
    # $B$+$D!"%H!<%/%sNs$N$"$kE@$h$j8e$mB&$r8+$?>l9g(B
    # $B3+;O%?%0$N8D?t$,=*N;%?%0$N8D?t$h$jB?$$$3$H$O$"$jF@$J$$$N$G(B
    # $BCzEYNI$$!#(B
    def distance_min_est_tag(other,pl_both=true)
	d = 0
	d1 = 0
	d1 += 3.5 if other.bold != @bold
	d1 += 4.5 if other.typewriter != @typewriter
	d1 += 3.5 if other.italic != @italic
	d1 += 3.5 if other.strong_emphasis != @strong_emphasis
	unless (other.strong_emphasis == @strong_emphasis and
		@strong_emphasis)
	    d1 += 4.5 if other.emphasis != @emphasis
	end
	
	d2 = (other.underline - @underline).abs * 3.5

	if other.plain?
	    dd1 = d1+d2
	    dd2 = 4.5
	    d += (dd1 <= dd2 ? dd1 : dd2)
	elsif self.plain?
	    dd1 = d1+d2
	    dd2 = 4.5
	    if pl_both
		d += (dd1 <= dd2 ? dd1 : dd2)
	    else
		d += dd1
	    end
	else
	    dd1 = d1+d2
	    plain_context = context_modified_by_tag(self,'PL')
	    dd2 = self.distance_min_est_tag(plain_context) +
		plain_context.distance_min_est_tag(other)
	    d += (dd1 <= dd2 ? dd1 : dd2)
	end
	
	d += 3.5 if other.color != @color
	d += 3.5 if other.size != @size
	
	d
    end

    def plain?
	! (@bold || @typewriter || @italic || @strong_emphasis ||
	   @emphasis || @underline > 0)
    end
end

########################################################################
# utility
def context_modified_by_tag(context,tag)
    cont = context.dup
    case tag
    when 'B'
	cont.bold = true
    when 'EM'
	cont.emphasis = (! cont.emphasis)
    when 'S'
	cont.strong_emphasis = true
    when 'I'
	cont.italic = true
    when 'PL'
	cont.underline_reset
	cont.bold = false
	cont.emphasis = false
	cont.italic = false
	cont.strong_emphasis = false
	cont.typewriter = false
    when 'TT'
	cont.typewriter = true
    when 'U'
	cont.underline_add
    when /\A\d\Z/
	cont.size = tag.to_i
    when /\A[rgbcmykw]\Z/
	cont.color = tag[0]
    else
	raise "unknown tag: #{tag}"
    end
    cont
end

