#!/usr/bin/env ruby

require 'context'
require 'smlngscanner'
require 'str_io'
require 'smlngparser'

########################################################################
# optimizer
class Node
    def initialize(parent,index,tag,type)
	@parent,@index,@tag,@type = parent,index,tag,type
	if parent
	    @tagstack = parent.tagstack.dup
	    @search_depth = parent.search_depth
	    case type
	    when :OPEN
		@tagstack.push [tag, parent.context]
		@context = context_modified_by_tag parent.context, tag
		@str = "<#{tag}>"
		@overhead_estimated = @str.size*2+1 + parent.overhead_estimated
		@search_depth += 1
	    when :CLOSE
		@str = ""
		poptag = nil
		until poptag == tag
		    poptag,@context = @tagstack.pop
		    @str += "</#{poptag}>"
		    unless poptag
			raise 'bug! try to close not found tag.'
		    end
		    @search_depth += 1
		end
	    else
		raise ArgumentError,
		    "type should be :OPEN or :CLOSE, but: #{type}"
	    end
	    @overhead = parent.overhead + @str.length
	else
	    # root node
	    @context = Context.new
	    @tagstack = []
	    @overhead = 0
	    @str = ''
	    @overhead_estimated = 0
	    @search_depth = 0
	end
	@overhead_estimated ||= parent.overhead_estimated

	
    end
    attr_reader :str
    attr_reader :context
    attr_reader :parent,:index,:tag,:type
    attr_reader :tagstack
    attr_reader :overhead	# Υ <b>ʤ3</TT>ʤ5
    attr_reader :overhead_estimated # OPENλΤߡĥ
    attr_reader :search_depth	# õڤο
end

class SMLNGOptimizer
    def initialize(text)
	@text = text
	@tokens = nil
	@tag_length_orig = nil

	@shortest_path_leaf = nil

	@search_done = false

	@depth_cutoff = nil
    end
    attr_reader :search_done
    attr_reader :tag_length_orig

    def count_steps
	minstep = (@tokens.size-1) / 2

	d = 0
	prev_context = Context.new
	@tokens.size.times do |i|
	    token = @tokens[i]
	    context = token.context
	    d += prev_context.distance_min_est(context)
	    prev_context = context
	end
	d += prev_context.distance_min_est(Context.new)

	# @depth_cutoff = minstep+1
	@depth_cutoff = d
	# ͥõξ硢˲򤬤ǽȤ
	# иŪΨ: 1.1 
	#@depth_cutoff = (d * 1.1).floor
    end

    def markup_start
	@rootnode = Node.new(nil,0,nil,nil)
	until @shortest_path_leaf
	    if ENV['MORO_TEST']
		print "depth cutoff=#{@depth_cutoff}\n"
	    end
	    markup(@rootnode,0)
	    @depth_cutoff += 1
	end
    end

    def minimum_overhead
	if @shortest_path_leaf
	    @shortest_path_leaf.overhead
	else
	    @tag_length_orig
	end
    end

    def open_tag(parent,index,tag)
	p = parent
	while p and p.index == index
	    if p.type == :CLOSE and p.tag == tag
		# ĤФΥ򳫤ȤƤ
		return
	    end
	    p = p.parent
	end
	node = Node.new(parent,index,tag,:OPEN)
	if node.overhead_estimated > minimum_overhead
	    if ENV['MORO_TEST'].to_i > 1
		print "(#{index})"
	    end
	    return
	end
	markup(node,index)
    end

    def close_tag(parent,index,tag)
	p = parent
	while p and p.index == index
	    if p.type == :OPEN # and p.tag == tag
		# ФΥĤ褦ȤƤ
		return
	    end
	    p = p.parent
	end
#	unless parent.tagstack.find { |t,c| t == tag }
#	    # ƤʤΤĤʤ
#	    return
#	end
	node = Node.new(parent,index,tag,:CLOSE)
	markup(node,index)
    end

    def markup(parent,index)
	#Thread.pass

	#ȿ
	#return if parent.search_depth > @depth_cutoff

	# դȿ
	if parent.search_depth > @depth_cutoff and
		@shortest_path_leaf and
		parent.parent and parent.parent.search_depth < @depth_cutoff
	    return
	end
	
	old_context = parent.context
	token = @tokens[index]

	if ENV['MORO_TEST_MARKUP']
	    printf "markup: %s %d\n", parent.str || '*', index
	    if parent and ENV['MORO_TEST_MARKUP'].to_i > 1
		print_path_0(STDERR,parent)
		STDERR.print "\n"
	    end
	end

	if token.nil?		# leaf
	    print "token end.\n" if ENV['MORO_TEST_MARKUP']
	    leaf = parent
	    until leaf.tagstack.empty?
		leaf = Node.new(leaf,index,leaf.tagstack[0][0],:CLOSE)
	    end
	    if (@shortest_path_leaf.nil? or
		@shortest_path_leaf.overhead > leaf.overhead) and
		    @tag_length_orig >= leaf.overhead
		@shortest_path_leaf = leaf
		if ENV['MORO_TEST']
		    print "shorter path found: #{leaf.overhead} (#{leaf.search_depth})\n"
		end
	    else
		if ENV['MORO_TEST'].to_i > 1
		    print "<<#{leaf.overhead}>>"
		end
	    end
	    return
	end
	
	new_context = token.context
	if old_context == new_context
	    markup(parent,index+1)
	    return
	end

	pl_contest = context_modified_by_tag(old_context,'PL')
	old_distance = old_context.distance(new_context)
	pl_distance  = pl_contest.distance(new_context)
	pl_valuable = old_distance >= pl_distance

	if old_context.strong_emphasis != new_context.strong_emphasis
	    if ! old_context.strong_emphasis
		open_tag(parent,index,'S')
		if parent.tagstack.find { |t,c| t == 'PL' }
		    close_tag(parent,index,'PL')
		end
	    else
		close_tag(parent,index,'S')
		open_tag(parent,index,'PL') if pl_valuable
	    end
	end
	if old_context.italic != new_context.italic
	    if ! old_context.italic
		open_tag(parent,index,'I')
		if parent.tagstack.find { |t,c| t == 'PL' }
		    close_tag(parent,index,'PL')
		end
	    else
		close_tag(parent,index,'I')
		open_tag(parent,index,'PL') if pl_valuable
	    end
	end
	if old_context.bold != new_context.bold
	    if ! old_context.bold
		open_tag(parent,index,'B')
		if parent.tagstack.find { |t,c| t == 'PL' }
		    close_tag(parent,index,'PL')
		end
	    else
		close_tag(parent,index,'B')
		open_tag(parent,index,'PL') if pl_valuable
	    end
	end
	if old_context.typewriter != new_context.typewriter
	    if ! old_context.typewriter
		open_tag(parent,index,'TT')
		if parent.tagstack.find { |t,c| t == 'PL' }
		    close_tag(parent,index,'PL')
		end
	    else
		close_tag(parent,index,'TT')
		open_tag(parent,index,'PL') if pl_valuable
	    end
	end

	# EM
	if old_context.emphasis != new_context.emphasis
	    if parent.tagstack.find { |t,c| t == 'EM' }
		close_tag(parent,index,'EM')
	    end
	    open_tag(parent,index,'EM')
	    if ! new_context.emphasis
		open_tag(parent,index,'PL') if pl_valuable
	    end
	    if parent.tagstack.find { |t,c| t == 'PL' }
		close_tag(parent,index,'PL')
	    end
	end

	#
	if old_context.underline != new_context.underline
	    if old_context.underline < new_context.underline
		open_tag(parent,index,'U')
		if parent.tagstack.find { |t,c| t == 'PL' }
		    close_tag(parent,index,'PL')
		end
	    else
		close_tag(parent,index,'U')
		open_tag(parent,index,'PL') if pl_valuable
	    end
	end
	#
	if old_context.color != new_context.color
	    old_tag = old_context.color.chr
	    new_tag = new_context.color.chr
	    if ENV['MORO_TEST_MARKUP']
		print "old_color: #{old_tag}\n"
		print "new_color: #{new_tag}\n"
	    end
	    if old_context.color != Context::COLOR_UNKNOWN
		# </old_color>
		close_tag(parent,index,old_tag)
	    end
	    if new_context.color != Context::COLOR_UNKNOWN
		# <new_color>
		open_tag(parent,index,new_tag)
	    end
	end
	#
	if old_context.size != new_context.size
	    old_tag = old_context.size.to_s
	    new_tag = new_context.size.to_s
	    if ENV['MORO_TEST_MARKUP']
		print "old_size: #{old_tag}\n"
		print "new_size: #{new_tag}\n"
	    end
	    if old_context.size != Context::SIZE_UNKNOWN
		# </old_size>
		close_tag(parent,index,old_tag)
	    end
	    if new_context.size != Context::SIZE_UNKNOWN
		# <new_size>
		open_tag(parent,index,new_tag)
	    end
	end
    end

    def print_path_0(out,node)
	if parent = node.parent
	    print_path_0(out,parent)
	    if node.index != parent.index
		out.print @tokens[parent.index].text
	    end
	end
	out.print node.str
    end

    def print_path(out)
	if @shortest_path_leaf.nil?
	    raise "path not found!\n"
	end
	print_path_0(out,@shortest_path_leaf)
	if @shortest_path_leaf.index < @tokens.size
	    out.print @tokens[-1].text
	end
    end

    def markup0(index=0, length=0)
	if index == 0
	    old_context = Context.new
	else
	    old_context = @tokens[index-1].context
	end
	new_context = @tokens[index].context

	if old_context.bold != new_context.bold
	    if ! old_context.bold
		# <B>ޤ</PL>(PL)
	    else
		# </B>ޤ<PL>
	    end
	end
	if old_context.italic != new_context.italic
	    if ! old_context.italic
		# <I>ޤ</PL>(PL)
	    else
		# </I>ޤ<PL>
	    end
	end
	if old_context.typewriter != new_context.typewriter
	    if ! old_context.typewriter
		# <TT>ޤ</PL>(PL)
	    else
		# </TT>ޤ<PL>
	    end
	end
	if old_context.strong_emphasis != new_context.strong_emphasis
	    if ! old_context.strong_emphasis
		# <S>ޤ</PL>(PL)
	    else
		# </S>ޤ<PL>
	    end
	end

	# ES-E	:
	# ES- S	:
	# ES-	: </EM> or <PL>
	# E -ES	: ⤷ʤ?
	# E - S	: ⤷ʤ?
	# E -	: </EM> or <PL>
	#  S-ES	: ⤷ʤ?
	#  S-E	: <PL>?<EM>
	#   -ES	: ⤷ʤ?
	#   -E	: <EM>
	if old_context.emphasis_effective != new_context.emphasis_effective
	    # XXX -
	end
	
	if old_context.underline != new_context.underline
	    if old_context.underline < new_context
		# <U>ޤ</PL>
	    else
		# </U>+ޤ<PL><U>*
	    end
	end
	if old_context.color != new_context.color
	    if old_context.color == Context::COLOR_UNKNOWN
		# <new_color>
	    else
		# </old_color>?<new_color>?
	    end
	end
	if old_context.size != new_context.size
	    if old_context.size == Context::SIZE_UNKNOWN
		# <new_size>
	    else
		# </new_size>?<new_color>?
	    end
	end
    end
    
    def run(out)
	parser = SMLNGParser.new(@text)
	parser.parse
	@tokens = parser.tokens
	@tag_length_orig = parser.tag_length
	count_steps
	if ENV['MORO_TEST']
	    print_tokens(out)
	    print "input overhead: #{@tag_length_orig}\n"
	end
	markup_start
	@search_done = true
	print_path(out)
    end

    def print_forcibly(out)
	if @shortest_path_leaf
	    print_path(out)
	else
	    out.print @text
	end
    end

    def print_tokens(out)
	@tokens.each do |token|
	    c = token.context
	    out.print (c.bold ? 'B' : ' ')
	    out.print (c.emphasis ? 'E' : ' ')
	    out.print (c.strong_emphasis ? 'S' : ' ')
	    out.print (c.italic ? 'I' : ' ')
	    out.print (c.typewriter ? 'T' : ' ')
	    out.print (c.underline.to_s)
	    out.print (c.color == Context::COLOR_UNKNOWN ? ' ' : c.color.chr)
	    out.print (c.size == Context::SIZE_UNKNOWN ? ' ' : c.size.to_s)
	    out.print ' '
	    out.print (token.type == :SPACE ? 'S' : 'T')
	    out.print ' '
	    out.print '"',token.text.gsub("\n", "\\n"),'"'
	    out.print "\n"
	end
    end
end

########################################################################
# main
if __FILE__==$0
    # »()tlimit롣
    # ̵0Ȥ롣
    tlimit_str = ARGV.shift || ENV['TLIMIT']
    if tlimit_str
	tlimit = tlimit_str.to_i
    end
    tlimit ||= 0

    text = STDIN.read

    optimizer = SMLNGOptimizer.new(text)

    if tlimit > 0
	Thread.start {
	    # ڤˤʤꤽˤʤä顢ʸ򤽤Τޤ޽Ϥ
	    Thread.current.priority += 2
	    sleep tlimit - 5
	    Thread.critical = true
	    unless optimizer.search_done
		optimizer.print_forcibly(STDOUT)
		exit
	    end
	    Thread.critical = false
	}
    end

    if false
	buf = StrIO.new
	optimizer.run(buf)
	print buf, "\n"
    else
	optimizer.run(STDOUT)
	#STDOUT.print "\n"
    end
end
