/* SML/NG Scanner
 * written by MOROHOSHI Akihiko <moro@nii.ac.jp>
 * $Id: smlngscanner.c,v 1.3 2001/07/27 15:05:00 moro Exp $
 */

#include <ruby.h>

#define SPC	0x20
#define CR	0x0d
#define LF	0x0a
#define TAB	0x09

static VALUE cScanner;
static VALUE sym_TEXT;
static VALUE sym_SPACE;
static VALUE sym_OPENTAG;
static VALUE sym_CLOSETAG;
static VALUE id_str;
static VALUE id_pointer;
static VALUE id_tag_length;

static VALUE scanner_initialize(VALUE self,VALUE str) {
    Check_Type(str, T_STRING);
    rb_ivar_set(self, id_str, str);
    rb_funcall(str, rb_intern("freeze"), 0);
    rb_ivar_set(self, id_pointer, INT2FIX(0));
    rb_ivar_set(self, id_tag_length, INT2FIX(0));

    /* printf("initialize: str=\"%s\"\n", STR2CSTR(str)); */
}

static void add_taglen(VALUE self, int len) {
    int oldlen = FIX2INT(rb_ivar_get(self, id_tag_length));
    rb_ivar_set(self, id_tag_length, INT2FIX(oldlen+len));
}

static VALUE scanner_scan(VALUE self) {
    char *p, *p0, *p00, *str;
    int pointer;

    VALUE rb_str = rb_ivar_get(self, id_str);
    VALUE rb_pointer = rb_ivar_get(self, id_pointer);
    VALUE sym,value;
    
    Check_Type(rb_str, T_STRING);
    Check_Type(rb_pointer, T_FIXNUM);
    str = STR2CSTR(rb_str);
    pointer = FIX2INT(rb_pointer);
    p = p0 = p00 = str + pointer;

    if (*p == '\0')
	return Qnil;

    switch (*p) {
    case SPC:
    case CR:
    case LF:
    case TAB:
	while (*p == SPC || *p == CR || *p == LF || *p == TAB)
	    p++;
	sym = sym_SPACE;
	value = rb_str_new(p0, p-p0);
	break;

    case '<':
	p++;
	if (*p == '/') {
	    sym = sym_CLOSETAG;
	    p++;
	} else {
	    sym = sym_OPENTAG;
	}

	p0 = p;
	while (*p != '>') {
	    if (*p == '\0')
		rb_raise(rb_eRuntimeError, "unexpected EOF.");
	    else
		p++;
	}
	value = rb_str_new(p0, p-p0);
	p++;
	add_taglen(self,p-p00);
	break;

    default:
	while (*p != '<' && *p != '\0')
	    p++;
	sym = sym_TEXT;
	value = rb_str_new(p0, p-p0);
	break;
    }

    rb_ivar_set(self, id_pointer, INT2FIX(p-str));
    return rb_ary_new3(2, sym, value);
}

void Init_smlngscanner() {
    cScanner = rb_define_class("SMLNGScanner", rb_cObject);

    rb_define_method(cScanner, "initialize", scanner_initialize, 1);
    rb_define_method(cScanner, "scan", scanner_scan, 0);
    rb_define_attr(cScanner, "pointer", 1, 0);
    rb_define_attr(cScanner, "tag_length", 1, 0);
    /* rb_define_singleton_method(cScanner, "scan", scan, 1); */

    sym_TEXT	= ID2SYM(rb_intern("TEXT"));
    sym_SPACE	= ID2SYM(rb_intern("SPACE"));
    sym_OPENTAG	= ID2SYM(rb_intern("OPENTAG"));
    sym_CLOSETAG= ID2SYM(rb_intern("CLOSETAG"));

    id_str = rb_intern("@str");
    id_pointer = rb_intern("@pointer");
    id_tag_length = rb_intern("@tag_length");
}
